package edu.tum.motras.algorithms;

import java.io.IOException;
import java.util.Set;
import java.util.concurrent.TimeoutException;

import edu.tum.xmts.State;
import edu.tum.xmts.TransitionSystem.Type;

/**
 * The {@code Algorithm} interface standardises the execution of algorithms on
 * modal transition systems. The algorithm is implemented in the
 * {@link #apply()} function.
 * 
 * As the number of arguments and supported systems can vary from algorithm to
 * algorithm this information can be queried via the methods {@link #getArity()}
 * and {@link #getSupportedTypes()}.
 * 
 * It may be the case that an algorithm needs additional parameters, e.g. a
 * synchronisation alphabet, a time-limit, an upper bound. These can be passed
 * via the {@link #setConfiguration(String)} method. The format of the string is
 * dependent on the implementation of the algorithm.
 * 
 * @author Salomon Sickert
 * 
 * @param <T>
 *            The type of the return value of the algorithm.
 */
public interface Algorithm<T> {

	/**
	 * Apply the algorithm on the given states.
	 * 
	 * @param states
	 *            The input of the algorithm.
	 * @return The result of applying the algorithm.
	 * @throws IllegalArgumentException
	 *             An {@link IllegalArgumentException} is thrown whenever the
	 *             number of passed states is not equal to the number specified
	 *             by {@link #getArity()}.
	 * 
	 * @throws TimeoutException
	 *             A {@link TimeoutException} is thrown whenever an answer
	 *             cannot be found within the maximum execution time.
	 * 
	 * @throws IOException
	 */
	public T apply(State... states) throws IllegalArgumentException, TimeoutException, IOException;

	public int getArity();

	public Set<Type> getSupportedTypes();

	public boolean isConfigurable();

	public void setConfiguration(String conf);
}
