package edu.tum.motras.algorithms;

import static com.google.common.base.Preconditions.checkNotNull;
import static edu.tum.motras.algorithms.util.Util.checkArguments;

import java.util.Collection;
import java.util.Set;

import com.google.common.collect.HashBasedTable;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Table;

import edu.tum.motras.algorithms.util.StatePair;
import edu.tum.motras.algorithms.util.Util;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem.Type;

public class ConsistencyRelation extends FixedPointAlgorithm<StatePair> implements Algorithm<Set<StatePair>> {

	private Table<State, State, StatePair> lookup;

	@Override
	public Set<StatePair> apply(State... input) throws IllegalArgumentException {
		checkArguments(this, input);
		Set<StatePair> newRelation = Util.getReachables(input[0], input[1]);

		lookup = HashBasedTable.create();
		for (StatePair pair : newRelation) {
			lookup.put(pair.get(1), pair.get(2), pair);
		}

		run(newRelation);

		lookup = null;
		return newRelation;
	}

	@Override
	public int getArity() {
		return 2;
	}

	@Override
	public Set<Type> getSupportedTypes() {
		return ImmutableSet.of(Type.MTS, Type.DMTS);
	}

	@Override
	public String toString() {
		return "Consistency Relation";
	}

	@Override
	protected boolean check(StatePair element) {
		State state1 = element.get(1);
		State state2 = element.get(2);

		for (Set<Transition> leftMustTransitions : Util.getDisjunctiveMustTransitions(state1)) {
			boolean matched = false;

			if (leftMustTransitions.isEmpty()) {
				continue;
			}

			for (Transition leftTransition : leftMustTransitions) {
				Collection<Transition> rightMayTransitions = state2.getTransitions(leftTransition.getAction().getId());

				for (Transition rightTransition : rightMayTransitions) {
					StatePair pair = lookup.get(leftTransition.getState(), rightTransition.getState());

					checkNotNull(pair, leftTransition.getId() + " " + rightTransition.getId());

					if (relation.contains(pair)) {
						matched = true;
						break;
					}
				}

				if (matched) {
					break;
				}
			}

			if (!matched) {
				return false;
			}
		}

		for (Set<Transition> rightMustTransitions : Util.getDisjunctiveMustTransitions(state2)) {
			boolean matched = false;

			if (rightMustTransitions.isEmpty()) {
				continue;
			}

			for (Transition rightTransition : rightMustTransitions) {
				Collection<Transition> leftMayTransitions = state1.getTransitions(rightTransition.getAction().getId());

				for (Transition leftTransition : leftMayTransitions) {
					StatePair pair = lookup.get(leftTransition.getState(), rightTransition.getState());

					checkNotNull(pair);

					if (relation.contains(pair)) {
						matched = true;
						break;
					}
				}

				if (matched) {
					break;
				}
			}

			if (!matched) {
				return false;
			}
		}

		return true;
	}

	@Override
	public boolean isConfigurable() {
		return false;
	}

	@Override
	public void setConfiguration(String conf) {
		throw new UnsupportedOperationException("This algorithm cannot be configured");
	}
}
