package edu.tum.motras.algorithms.util;

import static com.google.common.base.Preconditions.checkArgument;

import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Set;

import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;
import com.google.common.collect.ImmutableSet;

import edu.tum.motras.algorithms.Algorithm;
import edu.tum.xmts.Action;
import edu.tum.xmts.Obligation;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;

public class Util {
	public static class ActionFilter implements Predicate<Transition> {

		private final String action;

		public ActionFilter(String action) {
			this.action = action;
		}

		@Override
		public boolean apply(Transition transition) {
			return transition.getAction().getId().equals(action);
		}
	}

	public enum RelationConstruction {
		CARTESIAN, REACHABLE
	}

	public static class TransitionFilter implements Predicate<Transition> {

		private final Set<State> contexts;
		private final Set<Action> actions;
		private final Set<State> states;

		public TransitionFilter(Set<State> contexts, Set<Action> actions, Set<State> states) {
			this.contexts = ImmutableSet.copyOf(contexts);
			this.actions = ImmutableSet.copyOf(actions);
			this.states = ImmutableSet.copyOf(states);
		}

		public TransitionFilter(State context, Action action, State state) {
			this.contexts = ImmutableSet.of(context);
			this.actions = ImmutableSet.of(action);
			this.states = ImmutableSet.of(state);
		}

		@Override
		public boolean apply(Transition transition) {
			return contexts.contains(transition.getContext()) && actions.contains(transition.getAction())
					&& states.contains(transition.getState());
		}
	}

	public static void checkArguments(Algorithm<?> algorithm, State... states) throws IllegalArgumentException {
		checkArgument(states.length == algorithm.getArity());

		for (State state : states) {
			checkArgument(algorithm.getSupportedTypes().contains(state.getContext().getType()));
		}
	}

	public static Set<StatePair> getCartesianProduct(State state1, State state2) {
		Set<StatePair> relation = new HashSet<>();

		for (State leftState : state1.getContext().states.getElements()) {
			for (State rightState : state2.getContext().states.getElements()) {
				relation.add(new StatePair(leftState, rightState));
			}
		}

		return relation;
	}

	public static Set<Set<Transition>> getDisjunctiveMustTransitions(State state) {
		Obligation obligation = state.getObligation();

		if (obligation instanceof Obligation.All) {
			return wrap(((Obligation.All) obligation).getMustTransitions());
		} else if (obligation instanceof Obligation.Must) {
			return wrap(((Obligation.Must) obligation).getMustTransitions());
		} else if (obligation instanceof Obligation.DisjunctiveMust) {
			return ((Obligation.DisjunctiveMust) obligation).getDisjunctiveMustTransitions();
		}

		throw new IllegalArgumentException(state.getId() + "'s obligation (" + obligation.getClass().getSimpleName()
				+ ") is not convertabile to disjunctive must transitions.");
	}

	public static Collection<Transition> getMustTransitions(State state) {
		Obligation obligation = state.getObligation();

		if (obligation instanceof Obligation.All) {
			return ((Obligation.All) obligation).getMustTransitions();
		} else if (obligation instanceof Obligation.Must) {
			return ((Obligation.Must) obligation).getMustTransitions();
		}

		throw new IllegalArgumentException(state.getId() + "'s obligation (" + obligation.getClass().getSimpleName()
				+ ") is not convertabile to must transitions.");
	}

	public static Collection<Transition> getMustTransitions(State state, String action) {
		return Collections2.filter(getMustTransitions(state), new ActionFilter(action));
	}

	public static Set<StatePair> getReachables(State state1, State state2) {
		Set<StatePair> relation = new LinkedHashSet<>();

		Deque<StatePair> waitingQueue = new LinkedList<>();
		waitingQueue.add(new StatePair(state1, state2));

		while (!waitingQueue.isEmpty()) {
			StatePair statePair = waitingQueue.removeFirst();

			if (!relation.add(statePair)) {
				continue;
			}

			State leftState = statePair.get(1);
			State rightState = statePair.get(2);

			for (Transition leftTransition : leftState.getTransitions()) {
				for (Transition rightTransition : rightState.getTransitions(leftTransition.getAction().getId())) {
					StatePair successor = new StatePair(leftTransition.getState(), rightTransition.getState());

					if (!relation.contains(successor)) {
						waitingQueue.addLast(successor);
					}
				}
			}
		}

		return relation;
	}

	private static <K> Set<Set<K>> wrap(Collection<K> objects) {
		Set<Set<K>> wrappedObjects = new HashSet<>();

		for (K object : objects) {
			wrappedObjects.add(Collections.singleton(object));
		}

		return wrappedObjects;
	}

	public static <T> T removeFirst(LinkedHashSet<T> set) {
		Iterator<T> iter = set.iterator();
		T element = iter.next();
		iter.remove();
		return element;
	}

	public static void removeUnreachableStates(TransitionSystem transitionSystem, State initialState) {
		LinkedHashSet<State> unreachableStates = new LinkedHashSet<>(transitionSystem.states.getElements());
		LinkedHashSet<State> unprocessed = new LinkedHashSet<>();

		unprocessed.add(initialState);

		while (!unprocessed.isEmpty()) {
			State state = removeFirst(unprocessed);
			unreachableStates.remove(state);

			for (Transition transition : state.getTransitions()) {
				State nextState = transition.getState();

				if (unreachableStates.contains(nextState) && !unprocessed.contains(nextState)) {
					unprocessed.add(nextState);
				}
			}
		}

		for (State unreachableState : unreachableStates) {
			transitionSystem.states.remove(unreachableState);
		}
	}

}
