package edu.tum.motras.benchmarker;

import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeoutException;
import java.util.function.Predicate;
import java.util.logging.Logger;

import org.apache.commons.math3.stat.descriptive.moment.Mean;
import org.apache.commons.math3.stat.descriptive.moment.StandardDeviation;

import com.google.gson.Gson;

import edu.tum.motras.algorithms.Algorithm;
import edu.tum.motras.algorithms.ModalRefinementQSat;
import edu.tum.motras.benchmarker.Benchmark.Pair;
import edu.tum.xmts.State;

public class BenchmarkSuiteExecutor {

	private static Logger logger = Logger.getLogger(BenchmarkSuiteExecutor.class.getName());

	public static <T> void runBenchmark(Algorithm<T> algorithm, List<State> systems, Predicate<T> resultChecker) {
		final int arity = algorithm.getArity();
		final List<Double> data = new ArrayList<>(systems.size() / arity);
		final List<Double> externalData = algorithm instanceof ModalRefinementQSat ? new ArrayList<>(systems.size()
				/ arity) : null;

		int failedOperations = 0;
		int totalOperations = 0;

		while (systems.size() >= arity) {
			State[] input = new State[arity];

			for (int i = 0; i < input.length; i++) {
				input[i] = systems.get(i);
			}

			/* Ensure that GC doesn't kick-in later */
			System.gc();

			long time = System.currentTimeMillis();

			try {
				if (!resultChecker.test(algorithm.apply(input))) {
					logger.severe("ResultChecker detected wrong result!");
				}

				data.add(1.0 * (System.currentTimeMillis() - time));

				if (externalData != null) {
					externalData
							.add(1.0 * ((ModalRefinementQSat) algorithm).getInternalSolver().getLastExecutionTime());
				}

			} catch (TimeoutException | IllegalArgumentException | IOException e) {
				failedOperations++;
				logger.severe(e.getMessage());
			}

			totalOperations++;

			for (int i = 0; i < input.length; i++) {
				systems.remove(0);
			}
		}

		analyseAndPrint(data);

		if (externalData != null) {
			w("QBF solving time: ");
			analyseAndPrint(externalData);
		}

		if (failedOperations > 0) {
			w("Failed: " + failedOperations + "/" + totalOperations);
		}
	}

	public static void runBenchmarkSuite(Algorithm<Boolean> algorithm, File suite, boolean combined) throws IOException {
		Gson gson = new Gson();

		w("\n");
		w("===============================================\n");
		w("Running Benchmark Suite: " + suite.getName() + "\n");
		w("===============================================\n");
		w("\n");

		for (File benchmark : suite.listFiles()) {

			if (benchmark.isHidden()) {
				continue;
			}

			Benchmark set = Benchmark.readFrom(benchmark);

			w("Benchmark Parameters: " + gson.toJson(set) + "\n");

			if (combined) {
				List<State> combination = flatten(set.getTrue());
				combination.addAll(flatten(set.getFalse()));

				w("Combined: ");
				runBenchmark(algorithm, combination, b -> true);
				w("\n");
			} else {
				w("True: \t");
				runBenchmark(algorithm, flatten(set.getTrue()), b -> b);
				w("\n");

				w("False: \t");
				runBenchmark(algorithm, flatten(set.getFalse()), b -> !b);
				w("\n");
			}
		}
	}

	private static void analyseAndPrint(List<Double> data) {
		double dataArray[] = new double[data.size()];

		for (int i = 0, n = data.size(); i < n; i++) {
			dataArray[i] = data.get(i);
		}

		Mean mean = new Mean();
		mean.setData(dataArray);

		StandardDeviation standardDeviation = new StandardDeviation();
		standardDeviation.setData(dataArray);

		// Convert to seconds.
		double meanValue = mean.evaluate() / 1000;
		double standardDeviationValue = standardDeviation.evaluate() / 1000;

		DecimalFormat format = new DecimalFormat("######.##");
		w(format.format(meanValue) + " s\t(±" + format.format(standardDeviationValue) + " s)\t");
	}

	private static List<State> flatten(List<Pair> pairs) {
		List<State> systems = new ArrayList<>(2 * pairs.size());

		for (Pair pair : pairs) {
			systems.add(pair.leftSystem.getInitialState());
			systems.add(pair.rightSystem.getInitialState());
		}

		return systems;
	}

	private static void w(String s) {
		System.out.print(s);
	}
}
