package edu.tum.motras.legacy.pmtsrc.util;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.List;
import java.util.concurrent.TimeoutException;
import java.util.logging.Logger;

import edu.tum.motras.algorithms.util.ExternalTool;
import edu.tum.motras.legacy.pmtsrc.util.formula.AbstractLiteralFactory;
import edu.tum.motras.legacy.pmtsrc.util.formula.Literal;
import edu.tum.motras.legacy.pmtsrc.util.formula.LiteralRange;
import edu.tum.motras.legacy.pmtsrc.util.formula.QuantifiedBoolean;
import edu.tum.motras.legacy.pmtsrc.util.formula.Quantifier;
import edu.tum.xmts.utils.XMTSReader;

public class QDIMACS extends ExternalTool<QuantifiedBoolean, Boolean> {

	// QSat Solver and I/O File
	private final File file;
	private static final Logger logger = Logger.getLogger(XMTSReader.class.getName());

	public enum QSatSolverType {
		quantor, QuBE, sKizzo011, sKizzo08, sSolve, yquaffle
	}

	public QDIMACS() throws IOException {
		this(LegacyConfiguration.config.SELECTED_BACKEND);
	}

	public QDIMACS(QSatSolverType p) throws IOException {
		super(p.toString());

		switch (p) {
		case quantor:
			addParameter("-n");
			addParameter("--time-limit=" + LegacyConfiguration.config.QSAT_TIMEOUT);
			file = null;
			break;

		case sKizzo08:
		case sKizzo011:
			addParameter("-dimacsout");

			//$FALL-THROUGH$
		default:
			file = File.createTempFile("pmts-", ".qcnf");
			file.deleteOnExit();
			addParameter(file.getAbsolutePath());
		}
	}

	public Boolean solve(QuantifiedBoolean formula) throws IOException, TimeoutException {
		List<List<Literal>> clauses = formula.getClauses();

		if (clauses.isEmpty()) {
			return Boolean.FALSE;
		}

		if (clauses.size() == 1 && clauses.get(0).isEmpty()) {
			return Boolean.TRUE;
		}

		Boolean result = run(formula);

		if (result == null) {
			throw new TimeoutException();
		}

		return result;
	}

	@Override
	protected Boolean readOutput(int exitValue, String output) {
		if (file != null) {
			file.delete();
		}

		if (output.contains("TRUE")) {
			return Boolean.TRUE;
		}

		if (output.contains("FALSE")) {
			return Boolean.FALSE;
		}

		if (output.length() < 6) {
			logger.warning(output);
			return null;
		}

		switch (output.charAt(6)) {
		case '1':
			return Boolean.TRUE;

		case '0':
			return Boolean.FALSE;

		default:
			logger.warning(output);
			return null;
		}
	}

	@Override
	protected void writeInput(Writer writer, QuantifiedBoolean input) throws IOException {
		if (file != null) {
			file.createNewFile();
			try (BufferedWriter fileWriter = new BufferedWriter(new FileWriter(file))) {
				writeFormula(input, fileWriter);
			}
		} else {
			writeFormula(input, writer);
		}
	}

	private static void writeFormula(QuantifiedBoolean formula, Writer w) throws IOException {
		final List<List<Literal>> clauses = formula.getClauses();
		final int variableCounter = AbstractLiteralFactory.initializedFactories.peekLast().getLiteralRange().upperBound;

		boolean previousQuantifier;

		/* Write Header */
		w.write("p cnf " + variableCounter + " " + clauses.size() + "\n");

		/* Write Quantifiers */
		w.write("e ");
		previousQuantifier = Quantifier.EXISTS;

		for (Quantifier quantifier : formula.getQuantifiers()) {
			if (previousQuantifier != quantifier.type) {
				w.write("0\n");
				previousQuantifier = quantifier.type;
				w.write(quantifier.type == Quantifier.EXISTS ? "e " : "a ");
			}

			LiteralRange r = quantifier.getVariableRange();

			for (int i = r.lowerBound; i < r.upperBound; i++) {
				w.write(i + " ");
			}
		}

		w.write("0\n");

		/* Write Matrix */
		for (List<Literal> clause : clauses) {
			for (Literal l : clause) {
				if (l.negated) {
					w.write("-");
				}

				w.write(l.variable + " ");
			}

			w.write("0\n");
		}

		w.flush();
	}

}