package edu.tum.motras.legacy.pmtsrc.util.formula;

import it.unimi.dsi.fastutil.ints.Int2IntMap;

import java.util.Deque;
import java.util.LinkedList;
import java.util.Map;

public abstract class AbstractLiteralFactory {

	public enum Domain {
		AUXILIARYVARIABLE, PARAMETER, PREFIXEDTRANSITION, RELATION, TRANSITION, GENERIC
	}

	/* Static / Shared */

	private static boolean initLock = false;
	public static Deque<AbstractLiteralFactory> initializedFactories = new LinkedList<>();

	public static void deinitializeAll() {
		initializedFactories = new LinkedList<>();
	}

	public static Literal getTrue() {
		return new Literal(true);
	}

	public static Literal getFalse() {
		return new Literal(false);
	}

	/* Object Data */

	public final Domain domain;

	private LiteralRange range;
	private int lowerBound;
	private int upperBound;

	protected AbstractLiteralFactory(Domain d) {
		domain = d;
		range = null;
		lowerBound = -1;
		upperBound = -1;
	}

	public void startInitialization() {
		if (initLock) {
			throw new IllegalStateException("Initialization already in progress");
		}

		if (range != null) {
			throw new IllegalStateException("LiteralFactory is already initialized");
		}

		initLock = true;

		if (!initializedFactories.isEmpty()) {
			upperBound = lowerBound = initializedFactories.peekLast().getLiteralRange().upperBound;
		} else {
			upperBound = lowerBound = 1;
		}
	}

	public void finishInitialization() {
		if (!initLock) {
			throw new IllegalStateException("No initialization is in progress");
		}

		if (range != null) {
			throw new IllegalStateException("LiteralFactory is already initialized");
		}

		range = new LiteralRange(lowerBound, upperBound);
		initializedFactories.addLast(this);

		initLock = false;
	}

	public LiteralRange getLiteralRange() {
		if (range == null) {
			throw new IllegalStateException("LiteralFactory isn't initialized");
		}

		return range;
	}

	protected int incUpperBound() {
		return upperBound++;
	}

	protected int getVariable(Int2IntMap m, int key) {
		int variable = m.get(key);

		if (variable == 0) {
			if (range != null) {
				throw new IllegalStateException(
						"Initialization already finished. Creation of new variables is not allowed (" + key + ")");
			}

			variable = incUpperBound();
			m.put(key, variable);
		}

		return variable;
	}

	protected <K> int getVariable(Map<K, Integer> m, K key) {
		Integer variable = m.get(key);
		int intVariable;

		if (variable == null) {
			if (range != null) {
				throw new IllegalStateException(
						"Initialization already finished. Creation of new variables is not allowed (" + key + ")");
			}

			intVariable = incUpperBound();
			m.put(key, new Integer(intVariable));
		} else {
			intVariable = variable.intValue();
		}

		return intVariable;
	}
}