package edu.tum.xmts;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.WeakHashMap;

import org.eclipse.jdt.annotation.NonNull;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;

import com.google.common.collect.Sets;

import edu.tum.xmts.formula.EvaluateVisitor;
import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.formula.Formula.Atom;
import edu.tum.xmts.formula.RemoveVisitor;
import edu.tum.xmts.formula.Visitor;

@NonNullByDefault
public abstract class Obligation {

	public static class All extends Obligation {

		public All(State context) {
			super(context);
		}

		@Override
		public Formula getFormula() {
			return Formula.and(createAtoms(context.getTransitions()));
		}

		public Collection<Transition> getMustTransitions() {
			return Collections.unmodifiableCollection(context.getTransitions());
		}

		@Override
		void remove(Parameter parameter) {
			return;
		}

		@Override
		void remove(Transition transition) {
			return;
		}

		@Override
		public @NonNull Iterator<Set<Transition>> getTranIterator() {
			return getTran().iterator();
		}

		@Override
		public Set<Set<Transition>> getTran() {
			return Collections.singleton((Set<Transition>) new HashSet<>(context.getTransitions()));
		}
	}

	public static class DisjunctiveMust extends Obligation {

		private Map<Set<Transition>, Formula> disjunctiveMustTransitions;

		public DisjunctiveMust(State context, Set<List<Transition>> disjunctiveMust) {
			super(context);
			checkNotNull(disjunctiveMust);

			disjunctiveMustTransitions = new HashMap<>(disjunctiveMust.size());

			for (Collection<Transition> group : disjunctiveMust) {
				disjunctiveMustTransitions.put(new HashSet<>(group), Formula.or(createAtoms(group)));
			}
		}

		public DisjunctiveMust(State context, Collection<Collection<Transition>> disjunctiveMust) {
			super(context);
			checkNotNull(disjunctiveMust);

			disjunctiveMustTransitions = new HashMap<>(disjunctiveMust.size());

			for (Collection<Transition> group : disjunctiveMust) {
				disjunctiveMustTransitions.put(new HashSet<>(group), Formula.or(createAtoms(group)));
			}
		}

		public Set<Set<Transition>> getDisjunctiveMustTransitions() {
			return Collections.unmodifiableSet(disjunctiveMustTransitions.keySet());
		}

		@Override
		public Formula getFormula() {
			return Formula.and(disjunctiveMustTransitions.values());
		}

		@Override
		void remove(Parameter parameter) {
			return;
		}

		@Override
		void remove(Transition transition) {
			Map<Set<Transition>, Formula> newDisjunctiveMustTransitions = new HashMap<>(disjunctiveMustTransitions);

			for (Set<Transition> transitions : disjunctiveMustTransitions.keySet()) {
				if (transitions.remove(transition)) {
					newDisjunctiveMustTransitions.put(transitions, Formula.or(createAtoms(transitions)));
				}
			}

			disjunctiveMustTransitions = newDisjunctiveMustTransitions;
		}

		@Override
		public Set<Set<Transition>> getTran() {
			return Sets.newHashSet(getTranIterator());
		}

		@Override
		public Iterator<Set<Transition>> getTranIterator() {
			return new TranIterator(getContext().getTransitions(), getFormula());
		}
	}

	public static class Generic extends Obligation {

		private Formula formula;

		public Generic(State context, Formula formula) {
			super(context);
			checkNotNull(formula);
			this.formula = formula;
		}

		@Override
		public Formula getFormula() {
			return formula;
		}

		@Override
		void remove(Parameter parameter) {
			formula = formula.accept(new RemoveVisitor(parameter));
		}

		@Override
		void remove(Transition transition) {
			Visitor<Formula> visitor = new RemoveVisitor(transition);
			formula = formula.accept(visitor);
		}

		@Override
		public Set<Set<Transition>> getTran() {
			if (formula == Formula.TRUE) {
				return Sets.powerSet(new HashSet<>(context.getTransitions()));
			}

			if (formula == Formula.FALSE) {
				return Collections.emptySet();
			}

			return Sets.newHashSet(getTranIterator());
		}

		@Override
		public Iterator<Set<Transition>> getTranIterator() {
			return new TranIterator(getContext().getTransitions(), formula);
		}
	}

	public static class Must extends Obligation {

		private final Map<Transition, Formula> mustTransitions;

		public Must(State context, Collection<Transition> must) {
			super(context);
			checkNotNull(must);
			checkArgument(context.getTransitions().containsAll(must));

			mustTransitions = new WeakHashMap<>(must.size());

			for (Transition transition : must) {
				mustTransitions.put(transition, new Atom<>(transition));
			}
		}

		@Override
		public Formula getFormula() {
			return Formula.and(mustTransitions.values());
		}

		public Set<Transition> getMustTransitions() {
			return Collections.unmodifiableSet(mustTransitions.keySet());
		}

		@Override
		void remove(Parameter parameter) {
			return;
		}

		@Override
		void remove(Transition transition) {
			mustTransitions.remove(transition);
		}

		@Override
		public Set<Set<Transition>> getTran() {
			Set<Set<Transition>> tran = new HashSet<>();

			for (Set<Transition> trans : Sets.powerSet(Sets.difference(new HashSet<>(context.getTransitions()),
					getMustTransitions()))) {
				tran.add(Sets.union(trans, getMustTransitions()).immutableCopy());
			}

			return tran;
		}

		/* The collections library is missing the necessary null-annotations */
		@SuppressWarnings("null")
		@Override
		public Iterator<Set<Transition>> getTranIterator() {
			return getTran().iterator();
		}
	}

	public static class Tran extends Obligation {

		private final Set<Set<Transition>> tran;

		public Tran(State context, Set<Set<Transition>> tran) {
			super(context);
			this.tran = tran;
		}

		@Override
		public Formula getFormula() {
			List<Formula> formulas = new LinkedList<>();

			for (Set<Transition> transitions : tran) {
				List<Formula> group = new LinkedList<>();

				for (Transition transition : transitions) {
					group.add(new Atom<>(transition));
				}

				for (Transition transition : Sets.difference(new HashSet<>(context.getTransitions()), transitions)) {
					group.add(new Atom<>(transition).not());
				}

				formulas.add(Formula.and(group));
			}

			return Formula.or(formulas);
		}

		/* The collections library is missing the necessary null-annotations */
		@SuppressWarnings("null")
		@Override
		public Set<Set<Transition>> getTran() {
			return Collections.unmodifiableSet(tran);
		}

		/* The collections library is missing the necessary null-annotations */
		@SuppressWarnings("null")
		@Override
		public Iterator<Set<Transition>> getTranIterator() {
			return tran.iterator();
		}

		@Override
		void remove(Parameter parameter) {
			return;
		}

		@Override
		void remove(Transition transition) {
			for (Set<Transition> transitions : tran) {
				transitions.remove(transition);
			}
		}
	}

	private static class TranIterator implements Iterator<Set<Transition>> {

		private final Iterator<Set<Transition>> assignmentIterator;
		private @Nullable Set<Transition> next;
		private final Formula formula;

		/* The collections library is missing the necessary null-annotations */
		@SuppressWarnings("null")
		protected TranIterator(Collection<Transition> domain, Formula formula) {
			this.formula = formula;
			this.assignmentIterator = Sets.powerSet(new HashSet<>(domain)).iterator();
			this.next = Collections.emptySet();
			next();
		}

		@Override
		public boolean hasNext() {
			return (next != null);
		}

		@Override
		public Set<Transition> next() {
			Set<Transition> current = this.next;

			if (current == null) {
				throw new NoSuchElementException();
			}

			this.next = null;

			while (assignmentIterator.hasNext()) {
				this.next = assignmentIterator.next();
				EvaluateVisitor visitor = new EvaluateVisitor(next);

				if (formula.accept(visitor)) {
					break;
				}

				this.next = null;
			}

			return current;
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}
	}

	protected static List<Formula> createAtoms(Collection<Transition> transitions) {
		List<Formula> list = new LinkedList<>();

		for (Transition transition : transitions) {
			checkNotNull(transition);
			list.add(new Atom<>(transition));
		}

		return list;
	}

	protected final State context;

	protected Obligation(State context) {
		this.context = context;
	}

	public abstract Formula getFormula();

	public abstract Set<Set<Transition>> getTran();

	public abstract Iterator<Set<Transition>> getTranIterator();

	abstract void remove(Parameter parameter);

	abstract void remove(Transition transition);

	public State getContext() {
		return context;
	}
}