package edu.tum.xmts;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkState;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;

import edu.tum.xmts.formula.Formula;
import edu.tum.xmts.formula.RemoveVisitor;
import edu.tum.xmts.utils.XMTSReader;
import edu.tum.xmts.utils.XMTSReader.MalformedInputException;
import edu.tum.xmts.utils.XMTSWriter;

@NonNullByDefault
public class TransitionSystem extends GenericRecord<Object> implements Cloneable {

	public final class ActionPool extends Pool<Action> {
		@Override
		public Action create() {
			return create(generateId("a"));
		}

		@Override
		public Action create(String id) {
			checkArgument(isUniqueId(id), "%s is not a unique id.", id);

			Action a = new Action(TransitionSystem.this, id);
			actions.add(a);
			propertyChangeSupport.firePropertyChange(PROP_ACTIONS, null, a);
			return a;
		}

		@Override
		protected void removeHelper(Action a) {
			for (State state : states.getElements()) {
				Deque<Transition> removeTransitions = new LinkedList<>();

				for (Transition transition : state.getTransitions()) {
					if (transition.getAction() == a) {
						removeTransitions.add(transition);
					}
				}

				for (Transition transition : removeTransitions) {
					state.removeTransition(transition);
				}
			}

			propertyChangeSupport.firePropertyChange(PROP_ACTIONS, a, null);
		}
	}

	public final class ParameterPool extends Pool<Parameter> {
		@Override
		public Parameter create() {
			return create(generateId("p"));
		}

		@Override
		public Parameter create(String id) {
			checkState(isSupportedProperty(PROP_PARAMETERS));
			checkArgument(isUniqueId(id));

			Parameter p = new Parameter(TransitionSystem.this, id);
			parameters.add(p);
			propertyChangeSupport.firePropertyChange(PROP_PARAMETERS, null, p);
			return p;
		}

		@Override
		protected void removeHelper(Parameter p) {
			RemoveVisitor visitor = new RemoveVisitor(p);
			for (Action action : actions.getElements()) {
				Formula formula = action.getRequirement();

				if (formula != null) {
					action.setRequirement(formula.accept(visitor));
				}
			}

			for (State state : states.getElements()) {
				state.getObligation().remove(p);
			}

			propertyChangeSupport.firePropertyChange(PROP_PARAMETERS, p, null);
		}
	}

	/*
	 * Trade-Off: Use ArrayList instead of LinkedList, as get() should be faster
	 * than remove()
	 */
	public abstract class Pool<T extends GenericRecord<?>> implements PropertyChangeListener, Iterable<T> {
		private List<T> elements = new ArrayList<>();
		private Map<String, T> lookup = new HashMap<>();

		public boolean contains(String id) {
			return lookup.containsKey(id);
		}

		abstract public T create();

		abstract public T create(String id);

		/* The collections library is missing the necessary null-annotations */
		@SuppressWarnings("null")
		public List<T> getElements() {
			return Collections.unmodifiableList(elements);
		}

		/* The collections library is missing the necessary null-annotations */
		@SuppressWarnings("null")
		@Override
		public Iterator<T> iterator() {
			return Collections.unmodifiableList(elements).iterator();
		}

		public T lookup(@Nullable String id) {
			T object = lookup.get(id);

			if (object == null) {
				throw new NoSuchElementException("Non existing object referenced (" + id + ")");
			}

			return object;
		}

		@Override
		public void propertyChange(@Nullable PropertyChangeEvent evt) {
			if (evt == null) {
				return;
			}

			if (evt.getPropertyName().equals(PROP_ID)) {
				String newId = (String) evt.getNewValue();

				if (newId == null) {
					throw new RuntimeException("newId is null");
				}

				T element = lookup.remove(evt.getOldValue());
				lookup.put(newId, element);
			}
		}

		public void remove(T element) {
			elements.remove(element);
			lookup.remove(element.getId());
			removeHelper(element);

			element.removePropertyChangeListener(this);
			for (PropertyChangeListener listener : subscribedListener) {
				element.removePropertyChangeListener(listener);
			}
		}

		protected void add(T element) {
			elements.add(element);
			lookup.put(element.getId(), element);

			element.addPropertyChangeListener(this);
			for (PropertyChangeListener listener : subscribedListener) {
				element.addPropertyChangeListener(listener);
			}
		}

		abstract protected void removeHelper(T element);
	}

	public final class StatePool extends Pool<State> {

		@Override
		public State create() {
			return create(generateId("s"));
		}

		@Override
		public State create(String id) {
			checkArgument(isUniqueId(id));

			State s = new State(TransitionSystem.this, id);
			states.add(s);
			propertyChangeSupport.firePropertyChange(PROP_STATES, null, s);
			return s;
		}

		@Override
		protected void removeHelper(State s) {
			for (State state : states.getElements()) {
				Deque<Transition> removeTransitions = new LinkedList<>();

				for (Transition transition : state.getTransitions()) {
					if (transition.getState() == s) {
						removeTransitions.add(transition);
					}
				}

				for (Transition transition : removeTransitions) {
					state.removeTransition(transition);
				}
			}

			if (getInitialState() == s) {
				setInitialState(null);
			}

			propertyChangeSupport.firePropertyChange(PROP_STATES, s, null);
		}
	}

	public static enum Type {
		BMTS("bmts"), DMTS("dmts"), MTS("mts"), MTSD("mtsd"), MTSD_DPS("mtsd+dps"), PMTS("pmts");

		private final String name;

		private Type(String name) {
			this.name = name;
		}

		@Override
		public String toString() {
			return name;
		}
	}

	/* XMTS Properties */

	public static final String PROP_TYPE = "type";
	public static final String PROP_INITIAL_STATE = "initial_state";
	public static final String PROP_PARAMETERS = "parameters";
	public static final String PROP_ACTIONS = "actions";
	public static final String PROP_STATES = "states";

	private Type type;
	private @Nullable State initialState;

	public final ParameterPool parameters;
	public final ActionPool actions;
	public final StatePool states;

	final Collection<PropertyChangeListener> subscribedListener;

	public TransitionSystem(String id) {
		super(new Object(), id);

		type = Type.MTS;
		parameters = new ParameterPool();
		actions = new ActionPool();
		states = new StatePool();

		subscribedListener = new LinkedList<>();
	}

	public void clearInitialState() {
		setInitialState(null);
	}

	@Override
	public TransitionSystem clone() {
		XMTSReader reader = new XMTSReader();
		XMTSWriter writer = new XMTSWriter();

		try {
			return reader.read(writer.write(this));
		} catch (MalformedInputException e) {
			e.printStackTrace();
			throw new RuntimeException("XMTS-Library Internal Error");
		}
	}

	public @Nullable State getInitialState() {
		return initialState;
	}

	public Type getType() {
		return type;
	}

	@Override
	public boolean isSupportedProperty(String propName) {
		if (super.isSupportedProperty(propName)) {
			return true;
		}

		switch (propName) {
		case PROP_TYPE:
		case PROP_INITIAL_STATE:
		case PROP_ACTIONS:
		case PROP_STATES:
			return true;

		case PROP_PARAMETERS:
			return type == Type.MTSD_DPS || type == Type.PMTS;

		default:
			return false;
		}
	}

	public void setInitialState(@Nullable State initialState) {
		checkArgument(initialState == null || initialState.context == this);

		State oldInitialState = this.initialState;
		this.initialState = initialState;
		propertyChangeSupport.firePropertyChange(PROP_INITIAL_STATE, oldInitialState, initialState);
	}

	public void setType(Type type) {
		Type oldType = this.type;
		this.type = type;
		propertyChangeSupport.firePropertyChange(PROP_TYPE, oldType, type);
	}

	public void subscribe(PropertyChangeListener listener) {
		subscribedListener.add(listener);

		this.addPropertyChangeListener(listener);

		for (Parameter parameter : parameters) {
			parameter.addPropertyChangeListener(listener);
		}

		for (Action action : actions) {
			action.addPropertyChangeListener(listener);
		}

		for (State state : states) {
			state.addPropertyChangeListener(listener);

			for (Transition transition : state.getTransitions()) {
				transition.addPropertyChangeListener(listener);
			}
		}
	}

	public void unsubscribe(PropertyChangeListener listener) {
		subscribedListener.remove(listener);

		this.removePropertyChangeListener(listener);

		for (Parameter parameter : parameters) {
			parameter.removePropertyChangeListener(listener);
		}

		for (Action action : actions) {
			action.removePropertyChangeListener(listener);
		}

		for (State state : states) {
			state.removePropertyChangeListener(listener);

			for (Transition transition : state.getTransitions()) {
				transition.removePropertyChangeListener(listener);
			}
		}
	}

	protected String generateId(String prefix) {
		for (int i = 0; true; i++) {
			if (isUniqueId(prefix + i)) {
				return prefix + i;
			}
		}
	}

	boolean isUniqueId(String id) {
		return !parameters.contains(id) && !actions.contains(id) && !states.contains(id);
	}
}