package edu.tum.motras.algorithms.junit;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.Collection;
import java.util.Set;

import org.junit.Before;
import org.junit.Test;

import com.google.common.collect.ImmutableSet;

import edu.tum.motras.algorithms.Algorithm;
import edu.tum.motras.algorithms.Conjunction;
import edu.tum.motras.algorithms.ModalRefinementQSat;
import edu.tum.motras.algorithms.util.Util;
import edu.tum.xmts.State;
import edu.tum.xmts.Transition;
import edu.tum.xmts.TransitionSystem;
import edu.tum.xmts.TransitionSystem.Type;
import edu.tum.xmts.utils.XMTSGenerator;
import edu.tum.xmts.utils.XMTSGenerator.XMTSGeneratorConfiguration;
import edu.tum.xmts.utils.XMTSReader;

public class ConjunctionTest {

	private Algorithm<Boolean> modalRefinement;
	private Conjunction conjunction;

	@Before
	public void setUp() throws Exception {
		conjunction = new Conjunction();
		modalRefinement = new ModalRefinementQSat();
	}

	@Test
	public void testApply() throws Exception {
		String xmts1 = "mts test1[action a action b action c " + "state s1 [obligation all  a->s2]"
				+ "state s2 [obligation true b->s1 c->s1]]";

		String xmts2 = "mts test2[action a action b action c " + "state s1 [obligation true a->s2 a->s3]"
				+ "state s2 [obligation all  b->s1]" + "state s3 [obligation all  c->s1]]";

		XMTSReader reader = new XMTSReader();
		TransitionSystem system1 = reader.read(xmts1);
		TransitionSystem system2 = reader.read(xmts2);

		Collection<Transition> aTran;
		Collection<Transition> bTran;
		Collection<Transition> cTran;

		Set<Set<Transition>> disjMusts;
		Set<Transition> disjMust;

		State s1_s1 = conjunction.apply(system1.states.lookup("s1"), system2.states.lookup("s1"));

		/* Test TransitionSystem */
		TransitionSystem conjunctionSystem = s1_s1.getContext();
		assertEquals(Type.DMTS, conjunctionSystem.getType());
		assertEquals(conjunctionSystem.parameters.getElements().toString(), 0, conjunctionSystem.parameters
				.getElements().size());
		assertEquals(conjunctionSystem.actions.getElements().toString(), 3, conjunctionSystem.actions.getElements()
				.size());
		assertEquals(conjunctionSystem.states.getElements().toString(), 3, conjunctionSystem.states.getElements()
				.size());

		/* Test State {s1,s1} */
		aTran = s1_s1.getTransitions("a");
		bTran = s1_s1.getTransitions("b");
		cTran = s1_s1.getTransitions("c");

		assertEquals(aTran.toString(), 2, aTran.size());
		assertEquals(bTran.toString(), 0, bTran.size());
		assertEquals(cTran.toString(), 0, cTran.size());

		disjMusts = Util.getDisjunctiveMustTransitions(s1_s1);
		disjMust = disjMusts.iterator().next();

		assertEquals(disjMusts.toString(), 1, disjMusts.size());
		assertEquals(disjMust.toString(), 2, disjMust.size());
		assertTrue(disjMust.toString(), disjMust.containsAll(s1_s1.getTransitions()));

		Transition[] tmp = new Transition[2];
		disjMust.toArray(tmp);

		State s2_s2 = tmp[0].getState().getTransitions("b").isEmpty() ? tmp[1].getState() : tmp[0].getState();
		State s2_s3 = tmp[0].getState().getTransitions("c").isEmpty() ? tmp[1].getState() : tmp[0].getState();

		/* Test State {s2,s2} */
		aTran = s2_s2.getTransitions("a");
		bTran = s2_s2.getTransitions("b");
		cTran = s2_s2.getTransitions("c");

		assertEquals(aTran.toString(), 0, aTran.size());
		assertEquals(bTran.toString(), 1, bTran.size());
		assertEquals(cTran.toString(), 0, cTran.size());

		disjMusts = Util.getDisjunctiveMustTransitions(s2_s2);
		disjMust = disjMusts.iterator().next();

		assertEquals(disjMusts.toString(), 1, disjMusts.size());
		assertTrue(s2_s2.getTransitions().containsAll(disjMust));
		assertTrue(disjMust.containsAll(s2_s2.getTransitions()));
		assertEquals(s1_s1, disjMust.iterator().next().getState());

		/* Test State {s2,s3} */
		aTran = s2_s3.getTransitions("a");
		bTran = s2_s3.getTransitions("b");
		cTran = s2_s3.getTransitions("c");

		assertEquals(aTran.toString(), 0, aTran.size());
		assertEquals(bTran.toString(), 0, bTran.size());
		assertEquals(cTran.toString(), 1, cTran.size());

		disjMusts = Util.getDisjunctiveMustTransitions(s2_s3);
		disjMust = disjMusts.iterator().next();

		assertEquals(disjMusts.toString(), 1, disjMusts.size());
		assertTrue(s2_s3.getTransitions().containsAll(disjMust));
		assertTrue(disjMust.containsAll(s2_s3.getTransitions()));
		assertEquals(s1_s1, disjMust.iterator().next().getState());

		assertTrue(modalRefinement.apply(s1_s1, system1.states.lookup("s1")));
		assertTrue(modalRefinement.apply(s1_s1, system2.states.lookup("s1")));
	}

	@Test
	public void testApplyRandom() throws Exception {
		TransitionSystem system1, system2;
		State state1, state2, state12;

		XMTSGenerator generator = new XMTSGenerator();
		XMTSGeneratorConfiguration configuration = new XMTSGeneratorConfiguration();

		configuration.type = Type.MTS;
		configuration.parameters = 0;
		configuration.actions = 2;

		configuration.clusterInterfaceNodes = 1;
		configuration.clusterNodeDegree = 0;
		configuration.clusterSize = 1;

		configuration.systemNodeDegree = 2;
		configuration.systemSize = 3;

		for (int i = 0; i < 10; i++) {
			system1 = generator.generate(configuration, 13l + 2 * i);
			system2 = generator.generate(configuration, 13l + 2 * i + 1);

			state1 = system1.states.getElements().get(0);
			state2 = system2.states.getElements().get(0);
			state12 = conjunction.apply(state1, state2);

			if (state12 == null) {
				continue;
			}

			assertTrue(modalRefinement.apply(state12, state2));
			assertTrue(modalRefinement.apply(state12, state1));
		}
	}

	@Test
	public void testGetArity() {
		AlgorithmTestUtil.testGetArity(conjunction, 2);
	}

	@Test
	public void testGetSupportedTypes() {
		AlgorithmTestUtil.testGetSupportedTypes(conjunction, ImmutableSet.of(Type.MTS, Type.DMTS));
	}
}
